
package bia.moea;

import static bia.moea.NSGA2Knapsack.dataset;
import static bia.moea.NSGA2Knapsack.parameters;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

/**
 * This main class represents population and implements evolution process.
 *
 * @author Frantisek Hruska, Jiri Kubalik
 */
public class SGAKnapsack {

    //--- Input data.
    static File dataset;

    /**
     * Old and new population.
     */
    static KnapsackIndividual[] oldPop;
    static KnapsackIndividual[] newPop;

    /**
     * Newly generated individuals
     */
    static KnapsackIndividual[] offspringPop;

    /**
     * SGA parameters
     */
    public static KnapsackParameters parameters = KnapsackParameters.getInstance();


    /**
     * Main function.
     *
     * @param args the command line arguments
     */
    public static void main(String[] args) throws Exception {
        parameters.readConfigFile();
        
        dataset = new File(parameters.getDataset() + ".data");
        if (dataset.exists()) {
            importDatasetKnapsack();
        }
        else{
            System.out.println("Dataset file does not exist!");
            System.exit(0);
        }
        
        run();
    }

    public static void run() throws Exception {

        int cgen = 0;
        int totEvals = 0;

        // Allocate memory for arrays
        newPop = new KnapsackIndividual[parameters.getPopSize()];
        oldPop = new KnapsackIndividual[parameters.getPopSize()];
        offspringPop = new KnapsackIndividual[2];

        // Initialise population.
        totEvals = initializePopulation(oldPop);

        // Generate statistic data.
        KnapsackStatData.getStatistics(oldPop);
        KnapsackStatData.reportResults(totEvals);
        KnapsackStatData.printBest(oldPop);
        //--- Save the initial best
        ArrayList<KnapsackIndividual> initialFront = new ArrayList<KnapsackIndividual>();
        initialFront.add(oldPop[KnapsackStatData.indbest]);

        // Evolve population using either generational or steady-state replacement strategy
        evolvePopulation(oldPop, totEvals);

        //--- Save the final best
        KnapsackStatData.getStatistics(oldPop);
        ArrayList<KnapsackIndividual> finalFront = new ArrayList<KnapsackIndividual>();
        finalFront.add(oldPop[KnapsackStatData.indbest]);
        
        //--- Visualize non-dominated fronts
        KnapsackViz.visualizeNonDominatedFronts(initialFront, finalFront);
    }

    /**
     * Evolves the @param population either using a generational or steady-state replacement strategy.
     *
     * @param population the population to be evolved.
     * @param totEvals   a number of fitness function evaluations that have been calculated
     *                   during an initialization phase.
     */
    public static void evolvePopulation(KnapsackIndividual[] population, int totEvals) {

        switch (parameters.getReplacementType()) {
            case 0:
                evolvePopGenerational(population, totEvals);
                break;
            case 1:
                evolvePopSteadyState(population, totEvals);
                break;
            default:
                System.err.println("ERROR - Bad value of replacementType variable. Check if its value is 0 or 1.");
                System.exit(1);
        }
    }

    /**
     * Evolves the @param population using a generational replacement strategy.
     *
     * @param population the population to be evolved.
     * @param totEvals   a number of fitness function evaluations that have been calculated
     *                   during an initialization phase.
     */
    public static void evolvePopGenerational(KnapsackIndividual[] population, int totEvals) {

        // While the termination criterion is not satisfied ...
        while (totEvals < parameters.getEvaluations()) {
            int chroms_generated = 0;

            // Elitism. Copy best individual from old population.
            newPop[chroms_generated++] = oldPop[KnapsackStatData.indbest].clone();

            // While new population is not filled...
            while (chroms_generated < parameters.getPopSize()) {
                int[] par = selection(oldPop);
                KnapsackIndividual parent1 = oldPop[par[0]].clone();
                KnapsackIndividual parent2 = oldPop[par[1]].clone();
                // Creates two new individuals.
                if (Math.random() < parameters.getPC()) {
                    offspringPop = parent1.crossover(parent2);
                    offspringPop[0].mutation();
                    offspringPop[1].mutation();
                } else {
                    offspringPop[0] = parent1;
                    offspringPop[1] = parent2;
                    offspringPop[0].mutation();
                    offspringPop[1].mutation();
                }
                // Calculates fitness of new individuals.
                offspringPop[0].getFitness();
                offspringPop[1].getFitness();
                totEvals += 2;  // Increase number of evaluations done.

                // Get offspring to new parent population.
                newPop[chroms_generated++] = offspringPop[0];
                if (chroms_generated < parameters.getPopSize())
                    newPop[chroms_generated++] = offspringPop[1];
            }

            // Swap oldPop and newPop
            KnapsackIndividual[] tempPop;
            tempPop = oldPop;
            oldPop = newPop;
            newPop = tempPop;

            // Generate statistic data.
            KnapsackStatData.getStatistics(oldPop);

            // Print results.
            KnapsackStatData.reportResults(totEvals);
            KnapsackStatData.printBest(oldPop);
        }
    }

    /**
     * Evolves the @param population using a steady-state replacement strategy.
     *
     * @param population the population to be evolved.
     * @param totEvals   a number of fitness function evaluations that have been calculated
     *                   during an initialization phase.
     */
    public static void evolvePopSteadyState(KnapsackIndividual[] population, int totEvals) {
    /* PUT YOUR CODE HERE */

        int generationCounter = 0;

        // While the termination criterion is not satisfied ...
        while (totEvals < parameters.getEvaluations()) {
            int replacementIndex;   //--- index of the individual to be replaced by a newly generated offspring

            int[] par = selection(oldPop);
            KnapsackIndividual parent1 = oldPop[par[0]].clone();
            KnapsackIndividual parent2 = oldPop[par[1]].clone();
            // Creates two new individuals
            if (Math.random() < parameters.getPC()) {
                offspringPop = parent1.crossover(parent2);  //--- create offspring by a crossover
                offspringPop[0].mutation();                 //--- mutate the offspring
                offspringPop[1].mutation();
            } else {
                offspringPop[0] = parent1;
                offspringPop[1] = parent2;
                offspringPop[0].mutation();     //--- just mutate the parents
                offspringPop[1].mutation();
            }
            // Calculates fitness of new individuals
            offspringPop[0].getFitness();
            offspringPop[1].getFitness();
            totEvals += 2;  // Increase number of evaluations done

            // Find an individual to be replaced by the first child and replace it
            replacementIndex = worstInPopulation(oldPop);
            oldPop[replacementIndex] = offspringPop[0];
            // Find an individual to be replaced by the second child and replace it
            replacementIndex = worstInPopulation(oldPop);
            oldPop[replacementIndex] = offspringPop[1];

            //--- Regular reports
            generationCounter++;
            if (generationCounter == oldPop.length / 2) {
                // Generate statistic data
                KnapsackStatData.getStatistics(oldPop);
                // Print results
                KnapsackStatData.reportResults(totEvals);
                KnapsackStatData.printBest(oldPop);
                //---
                generationCounter = 0;
            }
        }

        //--- Final report
        // Generate statistic data
        KnapsackStatData.getStatistics(oldPop);
        // Print results
        KnapsackStatData.reportResults(totEvals);
        KnapsackStatData.printBest(oldPop);
    }

    /**
     * Initialization of population.
     *
     * @return a number of fitness evaluations calculated during the population initialization
     */
    public static int initializePopulation(KnapsackIndividual[] population) {
        int count = 0;

        // Generate individuals in parent population and calculate their fitness values.
        for (int ind = 0; ind < parameters.getPopSize(); ind++) {
            population[ind] = new KnapsackIndividual();
            population[ind].getFitness(); // Evaluate fitness of new individual.
            count++;
        }

        return count;
    }

    /**
     * @return an index of the worst individual in the current population.
     */
    public static int worstInPopulation(KnapsackIndividual[] population) {
        int worst = 0;

        for (int ind = 1; ind < parameters.getPopSize(); ind++) {
            if (population[worst].fitness > population[ind].fitness)
                worst = ind;
        }

        return worst;
    }

    /**
     * Makes deep copy of individual in population p with index from and stores it to population p2 to index to.
     */
    public static void copyIndividual(KnapsackIndividual source, int to, KnapsackIndividual[] population) {
        population[to] = source;
    }

    /**
     * Tournament selection with number of individuals in tournament.
     *
     * @return index of winner individual.
     */
    public static int tournamentSelection(KnapsackIndividual[] p) {
        int index;
        int winnerIndex = (int) (Math.random() * (parameters.getPopSize())); // First member of tournament selection.
        double winnerFitness = p[winnerIndex].fitness;

        // Choose N individuals and select best one.
        // We chose randomly first individual 3 rows up, so there is only SGA.tournament_size-1 random individuals to check. 
        for (int individualIndex = 0; individualIndex < parameters.getTournamentSize() - 1; individualIndex++) {
            index = (int) (Math.random() * (parameters.getPopSize()));
            if (p[index].fitness > winnerFitness) {
                winnerFitness = p[index].fitness;
                winnerIndex = index;
            }
        }

        return winnerIndex;
    }

    /**
     * Performs selection method according selectionType value.
     *
     * @return array of two indexes selected by selection methods.
     */
    public static int[] selection(KnapsackIndividual[] p) {
        int[] chosen = new int[2];

        switch (parameters.getSelectionType()) {
            case 0:
                chosen[0] = tournamentSelection(p);
                chosen[1] = chosen[0];
                while (chosen[0] == chosen[1])
                    chosen[1] = tournamentSelection(p);
                break;
            default:
                System.err.println("ERROR - Bad value of selectionType variable. Check if its value is 0.");
                System.exit(1);
        }

        return chosen;
    }

    /**
     * Imports data from specified input file.
     * 
     */
    private static void importDatasetKnapsack() {
        BufferedReader br;
        String temp;

        try {
            br = new BufferedReader(new FileReader(dataset));
            temp = br.readLine();   //--- reads first line "knapsack problem specification (...)"
            String[] line = temp.split(",");
            temp = line[1];
            line = temp.split("\\s++");
            parameters.setNbItems(Integer.parseInt(line[1]));

            KnapsackIndividual.items = new KnapsackItem[parameters.getNbItems()];
            for(int j = 0; j < parameters.getNbItems(); j++){
                KnapsackIndividual.items[j] = new KnapsackItem();
            }
            for(int i = 0; i < 2; i++){
                temp = br.readLine();   //--- reads "="
                temp = br.readLine();   //--- reads second line "knapsack 1:"
                temp = br.readLine();   //--- reads third line "capacity: ..."
                line = temp.split("\\+");
                KnapsackIndividual.capacities[i] = Integer.parseInt(line[1]);
                for(int j = 0; j < parameters.getNbItems(); j++){
                    temp = br.readLine();   //--- reads "item1:"
                    temp = br.readLine();   //--- reads "weight: ..."
                    line = temp.split("\\+");
                    KnapsackIndividual.items[j].setWeight(i, Integer.parseInt(line[1]));
                    //---
                    temp = br.readLine();   //--- reads "profit: ..."
                    line = temp.split("\\+");
                    KnapsackIndividual.items[j].setProfit(i, Integer.parseInt(line[1]));
                    //--- set wpRatio
                    KnapsackIndividual.items[j].calculateWPRatio(i);
                }
            }
            br.close();

            //--- Calculate WeightProfitOrder
            KnapsackIndividual.WeightProfitOrder = new int[2][];     //--- one order for each knapsack
            KnapsackIndividual.WeightProfitOrder[0] = new int[parameters.getNbItems()];
            KnapsackIndividual.WeightProfitOrder[1] = new int[parameters.getNbItems()];
            for(int i = 0; i < parameters.getNbItems(); i++){
                KnapsackIndividual.WeightProfitOrder[0][i] = i;
                KnapsackIndividual.WeightProfitOrder[1][i] = i;
            }
            //--- Sort items in descending order, from the least valuable to the most valuable one
            for(int i = 0; i < parameters.getNbItems(); i++){
                //--- 1st knapsack
                for(int j = 0; j < parameters.getNbItems()-1-i; j++){
                    if(KnapsackIndividual.items[KnapsackIndividual.WeightProfitOrder[0][j]].getWPRatio(0) < KnapsackIndividual.items[KnapsackIndividual.WeightProfitOrder[0][j+1]].getWPRatio(0)){
                        int k = KnapsackIndividual.WeightProfitOrder[0][j];
                        KnapsackIndividual.WeightProfitOrder[0][j] = KnapsackIndividual.WeightProfitOrder[0][j+1];
                        KnapsackIndividual.WeightProfitOrder[0][j+1] = k;
                    }
                }
                //--- 2nd knapsack
                for(int j = 0; j < parameters.getNbItems()-1-i; j++){
                    if(KnapsackIndividual.items[KnapsackIndividual.WeightProfitOrder[1][j]].getWPRatio(1) < KnapsackIndividual.items[KnapsackIndividual.WeightProfitOrder[1][j+1]].getWPRatio(1)){
                        int k = KnapsackIndividual.WeightProfitOrder[1][j];
                        KnapsackIndividual.WeightProfitOrder[1][j] = KnapsackIndividual.WeightProfitOrder[1][j+1];
                        KnapsackIndividual.WeightProfitOrder[1][j+1] = k;
                    }
                }
            }
        } catch (IOException e) {
            System.out.println(e.getMessage());
        }
    }
}
